# ruff: noqa: UP007
# @omlish-lite
"""
TODO:
 - reliable pid retrieval
  - contents are *ignored*, just advisory
  - check double-check:
   - 1) get pid of flock holder
   - 2) get pidfd to that
   - 3) recheck current pid of flock holder == that pid
  - racy as to if it's a different actual process as initial check, just with same pid, but due to 'identity' / semantic
    meaning of the named pidfile the processes are considered equivalent
  - read_checked(), contextmanager
"""
import fcntl
import os
import signal
import typing as ta


##


class Pidfile:
    def __init__(
            self,
            path: str,
            *,
            inheritable: bool = True,
    ) -> None:
        super().__init__()

        self._path = path
        self._inheritable = inheritable

    @property
    def path(self) -> str:
        return self._path

    @property
    def inheritable(self) -> bool:
        return self._inheritable

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._path!r})'

    #

    _f: ta.TextIO

    def fileno(self) -> ta.Optional[int]:
        if hasattr(self, '_f'):
            return self._f.fileno()
        else:
            return None

    #

    def __enter__(self) -> 'Pidfile':
        fd = os.open(self._path, os.O_RDWR | os.O_CREAT, 0o600)

        try:
            if self._inheritable:
                os.set_inheritable(fd, True)

            f = os.fdopen(fd, 'r+')

        except Exception:
            try:
                os.close(fd)
            except Exception:  # noqa
                pass
            raise

        self._f = f
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    #

    def __getstate__(self):
        state = self.__dict__.copy()

        if '_f' in state:
            # self._inheritable may be decoupled from actual file inheritability - for example when using the manager.
            if os.get_inheritable(fd := state.pop('_f').fileno()):
                state['__fd'] = fd

        return state

    def __setstate__(self, state):
        if '_f' in state:
            raise RuntimeError

        if '__fd' in state:
            state['_f'] = os.fdopen(state.pop('__fd'), 'r+')

        self.__dict__.update(state)

    #

    def close(self) -> bool:
        if not hasattr(self, '_f'):
            return False

        self._f.close()
        del self._f
        return True

    def try_lock(self) -> bool:
        try:
            fcntl.flock(self._f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True

        except OSError:
            return False

    def ensure_locked(self) -> None:
        if not self.try_lock():
            raise RuntimeError('Could not get lock')

    #

    def write(self, pid: ta.Optional[int] = None) -> None:
        self.ensure_locked()

        if pid is None:
            pid = os.getpid()

        self._f.seek(0)
        self._f.truncate()
        self._f.write(f'{pid}\n')
        self._f.flush()

    def clear(self) -> None:
        self.ensure_locked()

        self._f.seek(0)
        self._f.truncate()

    def read(self) -> int:
        if self.try_lock():
            raise RuntimeError('Got lock')

        self._f.seek(0)
        return int(self._f.read())

    def kill(self, sig: int = signal.SIGTERM) -> None:
        pid = self.read()
        os.kill(pid, sig)



# ruff: noqa: UP007
# @omlish-lite
import contextlib
import os
import threading
import typing as ta
import weakref

from ...lite.check import check
from .pidfile import Pidfile


##


class _PidfileManager:
    """
    Manager for controlled inheritance of Pidfiles across forks.

    Not implemented as an instantiated class as there is no way to unregister at_fork listeners, and because Pidfiles
    may be pickled and there must be no possibility of accidentally unpickling and instantiating a new instance of the
    manager.
    """

    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    _lock: ta.ClassVar[threading.Lock] = threading.Lock()
    _installed: ta.ClassVar[bool] = False
    _pidfile_threads: ta.ClassVar[ta.MutableMapping[Pidfile, threading.Thread]] = weakref.WeakKeyDictionary()

    @classmethod
    def _before_fork(cls) -> None:
        cls._lock.acquire()

    @classmethod
    def _after_fork_in_parent(cls) -> None:
        cls._lock.release()

    @classmethod
    def _after_fork_in_child(cls) -> None:
        th = threading.current_thread()
        for pf, pf_th in list(cls._pidfile_threads.items()):
            if pf_th is not th:
                pf.close()
                del cls._pidfile_threads[pf]

        cls._lock.release()

    #

    @classmethod
    def _install(cls) -> None:
        check.state(not cls._installed)

        os.register_at_fork(
            before=cls._before_fork,
            after_in_parent=cls._after_fork_in_parent,
            after_in_child=cls._after_fork_in_child,
        )

        cls._installed = True

    @classmethod
    def install(cls) -> bool:
        with cls._lock:
            if cls._installed:
                return False

            cls._install()
            return True

    @classmethod
    @contextlib.contextmanager
    def inheritable_pidfile_context(
            cls,
            path: str,
            *,
            inheritable: bool = True,
            **kwargs: ta.Any,
    ) -> ta.Iterator[Pidfile]:
        """
        A contextmanager for creating and managing a Pidfile which will only be inherited by forks of the calling /
        creating thread.
        """

        check.arg(inheritable)

        cls.install()

        pf = Pidfile(
            path,
            inheritable=False,
            **kwargs,
        )

        with cls._lock:
            cls._pidfile_threads[pf] = threading.current_thread()
        try:

            with pf:
                os.set_inheritable(check.not_none(pf.fileno()), True)
                yield pf

        finally:
            with cls._lock:
                del cls._pidfile_threads[pf]


def open_inheritable_pidfile(path: str, **kwargs: ta.Any) -> ta.ContextManager[Pidfile]:
    return _PidfileManager.inheritable_pidfile_context(path, **kwargs)  # noqa



"""
TODO:
 - handle secrets (but they're strs..)
"""
import typing as ta

from .. import cached
from .. import check
from .. import collections as col


StrOrBytes: ta.TypeAlias = str | bytes

CanHttpHeaders: ta.TypeAlias = ta.Union[
    'HttpHeaders',

    ta.Mapping[str, str],
    ta.Mapping[str, ta.Sequence[str]],
    ta.Mapping[str, str | ta.Sequence[str]],

    ta.Mapping[bytes, bytes],
    ta.Mapping[bytes, ta.Sequence[bytes]],
    ta.Mapping[bytes, bytes | ta.Sequence[bytes]],

    ta.Mapping[StrOrBytes, StrOrBytes],
    ta.Mapping[StrOrBytes, ta.Sequence[StrOrBytes]],
    ta.Mapping[StrOrBytes, StrOrBytes | ta.Sequence[StrOrBytes]],

    ta.Sequence[tuple[str, str]],
    ta.Sequence[tuple[bytes, bytes]],
    ta.Sequence[tuple[StrOrBytes, StrOrBytes]],
]


class HttpHeaders:
    def __init__(self, src: CanHttpHeaders) -> None:
        super().__init__()

        if isinstance(src, HttpHeaders):
            check.is_(src, self)
            return

        # TODO: optimized storage, 'use-whats-given'
        lst: list[tuple[bytes, bytes]] = []
        if isinstance(src, ta.Mapping):
            for k, v in src.items():
                if isinstance(v, (str, bytes)):
                    lst.append((self._as_bytes(k), self._as_bytes(v)))
                else:
                    for e in v:
                        lst.append((self._as_bytes(k), self._as_bytes(e)))

        elif isinstance(src, (str, bytes)):  # type: ignore
            raise TypeError(src)

        elif isinstance(src, ta.Sequence):
            for t in src:
                if isinstance(t, (str, bytes)):
                    raise TypeError(t)

                k, v = t
                lst.append((self._as_bytes(k), self._as_bytes(v)))

        else:
            raise TypeError(src)

        self._lst = lst

    def __new__(cls, obj: CanHttpHeaders) -> 'HttpHeaders':
        if isinstance(obj, HttpHeaders):
            return obj

        return super().__new__(cls)

    #

    # https://github.com/pgjones/hypercorn/commit/13f385be7277f407a9a361c958820515e16e217e
    ENCODING: ta.ClassVar[str] = 'latin1'

    @classmethod
    def _as_bytes(cls, o: StrOrBytes) -> bytes:
        if isinstance(o, bytes):
            return o
        elif isinstance(o, str):
            return o.encode(cls.ENCODING)
        else:
            raise TypeError(o)

    #

    @cached.function
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({{{", ".join(repr(k) for k in self.single_str_dct)}}})'

    #

    @property
    def raw(self) -> ta.Sequence[tuple[bytes, bytes]]:
        return self._lst

    @classmethod
    def _as_key(cls, o: StrOrBytes) -> bytes:
        return cls._as_bytes(o).lower()

    @cached.property
    def normalized(self) -> ta.Sequence[tuple[bytes, bytes]]:
        return [(self._as_key(k), v) for k, v in self._lst]

    #

    @cached.property
    def multi_dct(self) -> ta.Mapping[bytes, ta.Sequence[bytes]]:
        return col.multi_map(self.normalized)

    @cached.property
    def single_dct(self) -> ta.Mapping[bytes, bytes]:
        return {k: v[0] for k, v in self.multi_dct.items() if len(v) == 1}

    @cached.property
    def strict_dct(self) -> ta.Mapping[bytes, bytes]:
        return col.make_map(self.normalized, strict=True)

    #

    @cached.property
    def strs(self) -> ta.Sequence[tuple[str, str]]:
        return tuple((k.decode(self.ENCODING), v.decode(self.ENCODING)) for k, v in self.normalized)

    @cached.property
    def multi_str_dct(self) -> ta.Mapping[str, ta.Sequence[str]]:
        return col.multi_map(self.strs)

    @cached.property
    def single_str_dct(self) -> ta.Mapping[str, str]:
        return {k: v[0] for k, v in self.multi_str_dct.items() if len(v) == 1}

    @cached.property
    def strict_str_dct(self) -> ta.Mapping[str, str]:
        return col.make_map(self.strs, strict=True)

    #

    def __bool__(self) -> bool:
        return bool(self._lst)

    def __len__(self) -> int:
        return len(self._lst)

    def __iter__(self) -> ta.Iterator[tuple[bytes, bytes]]:
        return iter(self._lst)

    @ta.overload
    def __getitem__(self, item: str) -> ta.Sequence[str]:
        ...

    @ta.overload
    def __getitem__(self, item: bytes) -> ta.Sequence[bytes]:
        ...

    @ta.overload
    def __getitem__(self, item: int) -> tuple[StrOrBytes, StrOrBytes]:
        ...

    @ta.overload
    def __getitem__(self, item: slice) -> ta.Sequence[tuple[StrOrBytes, StrOrBytes]]:
        ...

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return self._lst[item]
        elif isinstance(item, str):
            return self.multi_str_dct[item.lower()]
        elif isinstance(item, bytes):
            return self.multi_dct[self._as_key(item)]
        else:
            raise TypeError(item)

    def keys(self) -> ta.Iterable[bytes]:
        return self.multi_dct.keys()

    def items(self) -> ta.Iterable[tuple[bytes, bytes]]:
        return self._lst


def headers(src: CanHttpHeaders) -> HttpHeaders:
    return HttpHeaders(src)
